from enum import Enum, auto
class velocityUnits(Enum):
    kRPM = auto()
    kRPS = auto()
    kEncoderTicksPer100 = auto()

class positionUnits(Enum):
    kEncoderTicks = auto()
    kRotations = auto()



import wpilib
from enum import Enum
import random

class Mode(Enum):
    RAINBOWHUE = 0
    RAINBOWBRIGHT = 1
    BLINK = 2
    FLASH = 3
    STATIC = 4
    RANDOM = 5

class Strip():
    def __init__(self, strip: list[wpilib.AddressableLED.LEDData],
                 name: str,
                 mode: Mode = Mode.RAINBOWBRIGHT,
                 hsv: list[int, int, int] = [243, 255, 128],
                 rgbP: list[int, int, int] = [0, 255, 0],
                 rgbS: list[int,int,int] = [255,255,255]):
        self.strip = strip
        self.name = name
        self.hsv = hsv[:]
        self.rgbPrimary = rgbP[:]
        self.rgbSecondary = rgbS[:]
        self.mode = mode

        self.ledPeriodicRate = 1.0
        self.fadeRate = 8
        self.rateS = 1.0 #default LED effects 1/s
        self.dutyCycle = 0.5
        self.timer = wpilib.Timer()
        self.timer.start()

    def periodic(self):
        match self.mode:
            case Mode.RAINBOWHUE:
                self.hsv[0] = rainbowHue(self.strip, self.hsv[0], self.hsv[1], self.hsv[2])
            case Mode.RAINBOWBRIGHT:
                self.hsv[2] = rainbowValue(self.strip, self.hsv[0], self.hsv[1], self.hsv[2], self.fadeRate)
            case Mode.BLINK |  Mode.FLASH | Mode.STATIC:
                blink(self.strip, self.timer, self.rateS, self.dutyCycle, self.rgbPrimary, self.rgbSecondary)
            case Mode.RANDOM:
                randomLeds(self.strip, self.timer, self.rateS)


    def getDefaultHue(self):
        if wpilib.DriverStation.getAlliance() == wpilib.DriverStation.Alliance.kBlue:
            return 243 #HSV Blue
        else:
            return 360 #HSV Red
    def getDefaultRgb(self):
        if wpilib.DriverStation.getAlliance() == wpilib.DriverStation.Alliance.kBlue:
            return [0, 0, 255]
        else:
            return [255, 0, 0]

    def setRainbowHue(self, startingHue: float = 0, saturation: float = 255, value: float = 128):
        self.hue = startingHue
        self.saturation = saturation
        self.value = value
        self.mode = Mode.RAINBOWHUE

    def setRainbowValue(self, startingHue: float = 0, saturation: float = 255, value: float = 128):
        self.hue = startingHue
        self.saturation = saturation
        self.value = value
        self.mode = Mode.RAINBOWBRIGHT
    def setBlink(self, rgb: list[int, int, int], rateS = 1.0, dutyCycle: float = 0.5):
        self.timer.reset()
        self.rateS = rateS
        self.dutyCycle = dutyCycle
        self.rgbPrimary = rgb
        self.rgbSecondary = [0,0,0]
        self.mode = Mode.BLINK
    def setFlash(self, OnRgb: list[int, int, int], OffRgb: list[int, int, int], rateS = 1.0, dutyCycle: float = 0.5):
        self.timer.reset()
        self.rateS = rateS
        self.dutyCycle = dutyCycle
        self.rgbPrimary = OnRgb
        self.rgbSecondary = OffRgb
        self.mode = Mode.FLASH
    def setStatic(self, rgb: list[int,int,int]):
        self.dutyCycle = 1.0
        self.rateS = 1.0
        self.rgbPrimary = rgb
        self.mode = Mode.STATIC
    def setRandom(self, rateS: float):
        self.dutyCycle = 1.0
        self.rateS = rateS
        self.mode = Mode.RANDOM
    def setTeamColor(self, chase = True):
        if chase:
            self.setRainbowValue(self.getDefaultHue(), 255, 128)
        else:
            self.setStatic(self.getDefaultRgb())


def rainbowHue(leds: list[wpilib.AddressableLED.LEDData], currentHue: float, saturation: float = 255, value: float = 128):
    """
    Pass in the LEDs array and the current hue to animate a rainbow
    The new next hue is passed out
    """
    # For every pixel
    for i in range(len(leds)):
        # Calculate the hue - hue is easier for rainbows because the color
        # shape is a circle so only one value needs to precess
        hue = (currentHue + (i * 180 / len(leds))) % 180

        # Set the value
        leds[i].setHSV(int(hue), saturation, value)

    # Increase by to make the rainbow "move"
    currentHue += 3

    # Check bounds
    return currentHue % 180

def rainbowValue(leds: [wpilib.AddressableLED.LEDData], hue: int = 360, saturation: int = 255, startingValue: int = 0, step: float = 16):
    """
    Pass in the LEDs array and the current hue to animate a rainbow
    The new next hue is passed out
    """
    # For every pixel
    for i in range(len(leds)):
        # Calculate the hue - hue is easier for rainbows because the color
        # shape is a circle so only one value needs to precess
        value = (startingValue + (i * 180 / len(leds))) % 255

        # Set the value
        leds[i].setHSV(int(hue), int(saturation), int(value))

    # Increase by to make the rainbow "move"
    startingValue += step

    # Check bounds
    return startingValue % 255

def blink(leds: list[wpilib.AddressableLED.LEDData], timer: wpilib.Timer, rateS: float = 1.0, dutyCycle: float = 0.5, onRgb: list[int,int,int] = [0,255,0], offRgb: [int, int, int] = [0,0,0]):
    #check for rollover
    if timer.get() > rateS:
        timer.reset()

    #determine if should be on or off
    rgb = offRgb
    if timer.get() < rateS * dutyCycle:
        rgb = onRgb

    for led in leds:
        led.setRGB(*rgb)

def randomLeds(leds: list[wpilib.AddressableLED.LEDData], timer: wpilib.Timer, rateS: float = 1.0):
    if timer.get() < rateS:
        return
    timer.reset()

    for led in leds:
        led.setRGB(random.randint(0,255), random.randint(0,255), random.randint(0,255))



import logging
import importlib
import sys
from posixpath import basename
import re
import os
from pathlib import Path
import inspect
from raptacon3200.utils import botFinder as utils

# Easy Line Debugging
# https://github.com/gruns/icecream
from icecream import ic

# Imported just to define the object returned from the BotFactory
from robots.configBasedRobot import ConfigBasedCommandRobot


logging.basicConfig(level=logging.DEBUG)
log = logging.root.getChild("BotFactory")
log.setLevel(logging.DEBUG)


class BotFactory(object):
    robot_full_name = None  # nameBot"
    raw_name = None  # Whatever ws given to the Factory as a name
    robot_class_name = None  # NameBot

    def __init__(self, name: str = None) -> ConfigBasedCommandRobot:
        """BotFactory will get the configs and create the bot for the supplied name.
        If no name is supplied, then it will query the filesystem for ~/robotConfig and
        read the robot name from the file (yaml extension not necessary) Bot is assumed at the end of the name.
        Eventually this method should be able to query the RoboRio itself for the robot name but this still still unimplmented."

        Returns:
            ConfigBasedCommandRobot: Instance of the robot like greenBot, dumboBot, teapotBot, labBot, etc
        """
        botName = None
        # User supplied the name so use that
        if name is not None:
            log.info(f"Using supplied name {name} for bot")
            botName = name
        # We need to figure out which bot to create
        else:
            log.info("Determining name of bot")
            # Robo Rio Query
            # TODO what is the Pythonic way of doing this???
            try:
                ic()
                botName = self._get_name_from_robo_rio()
            except LookupError as e:
                ic(e)
                try:
                    botName = self._get_name_from_config_file()
                except LookupError as e:
                    ic(e)
                    raise LookupError(
                        f"----> Unable to determine bot name from anything. Please look at README.md and READ on how to properly define your config file {e}"
                    )

        # Remove any extension from the botName as this will be computed
        ic(f"before normalize {botName}")
        botName = self._normalize_bot_name(botName)
        ic(f"after normalize {botName}")

        # Validate the bot name and match it to a yaml file, verify the config and python exist for the bot too
        try:
            # TODO I believe we should parse, validate and fill out self.robotFullName, modulename etc
            self._validate_bot_name_for_yaml_file(botName)
            self._validate_bot_config_exists(botName)
            self._validate_bot_python_exists(botName)
        except Exception as e:
            raise AssertionError(
                "The " + botName + " could not be found or had errors. error=" + e
            )

        # If we made it here, we know our botName is valid and the python for the bot exists and the config for the bot exists

    def _normalize_bot_name(self, robotName: str) -> str:
        """Takes in a name of <robot name>Bot.py, <robot name>, <robot name>Bot <robot name>Bot.yml/yaml and
        will normalize to just <robot name> and then assign to internal vars for future use

        Args:
            name (str): The name of the robot in the form of <robot name> or <robot name>Bot.yml/yaml

        Returns:
            str: The normalize name of the robot in the form of <robot name> ie dumbo or lab
        """
        pos = robotName.find("Bot")
        ic(f"Bot={pos}")
        # If we have the format of <robot name> then just take the robotName in it's entirety
        if pos == -1:
            name = robotName
        else:
            name = robotName[:pos]
        ic(f"name={name}")

        self.raw_name = robotName
        self.name = name
        self.robot_full_name = f"{name}Bot"
        self.robot_class_name = f"{name.capitalize()}Bot"
        self.robot_module_name = f"robots.{self.robot_full_name}"

        return self.name

    def _validate_bot_name_for_yaml_file(self, filename: str) -> bool:
        """Checks for <name>Bot.yaml or <name>Bot.yml

        Args:
            filename (str): The name of the file to check for config

        Returns:
            bool: True if the file matches our desired syntax otherwise False
        """
        ic(f"validateBotNameForYamlFile before {filename}")
        filename = filename + "Bot.yml"
        ic(f"validateBotNameForYamlFile after {filename}")
        pattern = r"^\S+Bot\.(yml|yaml)$"
        return bool(re.match(pattern, filename))

    # Moved to botFinder.py
    # def _find_robot_in_stack(self) -> str:
    #     """Looks in call stack for first robot.py and returns the fully qualified path to that file

    #     Returns:
    #         str: Fully qualified path to robot.py from call stack (minus robot.py)
    #     """
    #     stack = inspect.stack()
    #     for f_idx, f_info in enumerate(stack):
    #         # print(f"robot_info stack {f_idx} {f_info}")
    #         frame, filename, line_number, function_name, lines, index = stack[f_idx]
    #         # print(
    #         #     f"looking for {f_idx} {f_info} {frame}, {filename}, {line_number}, {function_name}, {lines}, {index}"
    #         # )
    #         if "robot.py" in filename:
    #             print(f"FOUND!!! {filename}")
    #             return os.path.dirname(filename)
    #     return None

    def _validate_bot_config_exists(self, filename: str) -> bool:
        ic(f"validateBotConfigExist before {filename}")
        filename = filename + "Bot.yml"
        ic(f"validateBotNameForYamlFile after {filename}")
        ic(filename)
        path = utils.botFinder()
        if path == None:
            raise LookupError(
                f"Unable to determine where robot.py lives by looking in call stack. Bug?"
            )
        configPath = path + os.path.sep + "configs" + os.path.sep + filename
        ic(configPath)

        if not os.path.isfile(configPath):
            log.error(f"Could not find {configPath}")
            return False

        return True

    def _validate_bot_python_exists(self, botname: str) -> bool:
        """Checks to see if the python file exists for the supplied bot name. Will crawl the call stack looking
        for robot.py and then take the path and look for the bot itself in the robots/ directory

        Args:
            filename (str): The name of the bot to check for the python file

        Returns:
            bool: True if the python file for the bot was found or False otherwies
        """
        ic(f"validateBotPythonExist before {botname}")
        filename = botname + "Bot.py"
        ic(f"validateBotPythonFile after {filename}")
        ic(filename)

        path = utils.botFinder()
        if path == None:
            raise LookupError(
                f"Unable to determine where robot.py lives by looking in call stack. Bug?"
            )
        configPath = path + os.path.sep + "robots" + os.path.sep + filename

        # refactor stuff should go here instead look in botFactory.py_refactored_wrong_thing_home_is_find_config_location_is_not for how we did it originally
        # configPath = self.getConfigPath(filename)
        # configPath = (
        #     os.path.dirname(__file__)
        #     + os.path.sep
        #     + ".."
        #     + os.path.sep
        #     + "robots"
        #     + os.path.sep
        #     + filename
        # )
        ic(configPath)

        if not os.path.isfile(configPath):
            log.error(f"Could not find {configPath}")
            return False

        return True

    def getConfigPath(self, filename: str) -> str:
        ic(f"getConfigPath {filename}")
        configPath = ""

        # Attempt to determine the directory of where robot.py lives to auto-discover the configs
        stack = inspect.stack()
        for i in range(len(stack)):
            stackpath = stack[i].filename
            ic(f"i={i}", stackpath)
            filename = os.path.basename(stackpath)
            ic(f"filename={filename}")
            if filename == "robot.py":
                ic(f"FOUND ROBOT.PY")
                configPath = os.path.dirname(stackpath)
                ic(f"configPath={configPath}")
                break
        if not configPath:
            raise LookupError(
                f"Unable to determine where robot.py lives. This is PROBABLY an error or bug."
            )

        fullyQualifiedConfigPath = (
            configPath + os.path.sep + "configs" + os.path.sep + filename
        )
        ic(fullyQualifiedConfigPath)
        ic(">>>>>>>>>>>>>>>>>>>>>>>>>.")
        ic()
        # for some reason this fullyqualifyed config path is returning robot.py instead of proper filename?
        return fullyQualifiedConfigPath

    def _get_name_from_config_file(self) -> str:
        """Locates a config file in ~/robotConfig and if found, will read the name of bot

        Raises:
            LookupError: If the fully qualified robotConfig file can't be opened
            LookupError: If the robotConfig can't be read

        Returns:
            str: The stripped contents of the config file
        """
        ic()
        # Need to supply a sane value to findConfig in the form of <botname>Bot.yml

        home = str(Path.home()) + os.path.sep
        robotConfigFile = home + "robotConfig"

        if not os.path.isfile(robotConfigFile):
            raise LookupError(
                f"Unable to open configFile.  {robotConfigFile} not found."
            )

        # We can find the file so lets open it and attempt to determine the bot name
        with open(robotConfigFile) as file:
            name = file.read()
            if name is None:
                raise LookupError(
                    f"Unable to determine name from configFile.  {robotConfigFile} is empty."
                )

            # Remove any leading and trailing whitespace in the file
            name = name.strip()
            ic(name)

            return name

    def _get_name_from_robo_rio(self) -> str:
        """Gets the robot name from the roboRio

        Raises:
            LookupError: If name can't be determined raise a lookup error

        Returns:
            str: name of the bot stored in the roboRio
        """
        ic()
        raise LookupError("Unable to determine name from roboRio. NOT IMPLEMENTED")

    def get_robot_name(self) -> str:
        """Determine the name of the robot from either the config file or the RoboRio memory

        Returns:
            str: name of the robot like green, dumbo, teapot, lab, etc
        """
        return "greenbot"

    def get_robot(self) -> ConfigBasedCommandRobot:
        # The name should be greenBot or dumboBot (no yaml/yml/py)
        ic(f"getRobotFromName {self.name}")
        # TODO Need an interator that does this to all bots
        # from robots.name import selfDumbo
        # robots.dumbo
        lib = f"robots.{self.robot_full_name}"

        ic(lib)
        # module = importlib.import_module(lib)
        module = None
        ic(self.robot_module_name)
        try:
            robots_path = utils.botFinder() + os.path.sep + "robots" + os.path.sep
            ic(robots_path)
            fully_qualified_robot = robots_path + "breadBoxBot.py"
            ic(fully_qualified_robot)
            # From https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
            # Do an absolute import since we're not generic and really have no idea what we're relative to
            spec = importlib.util.spec_from_file_location(lib, fully_qualified_robot)
            ic(spec)
            module = importlib.util.module_from_spec(spec)
            ic(module)
            sys.modules[lib] = module
            # module = importlib.import_module(robots_path + lib)
            spec.loader.exec_module(module)
            # foo.MyClass()

        except Exception as e:
            err_msg = f"Unable to import module {fully_qualified_robot} please ensure that your ~/robotConfig exists, and contents are correct. robot name from config file={self.raw_name} Exception: {e}"
            log.error(err_msg)
            raise ImportError(err_msg)
        ic(module)
        # Dynamically get the robot_class_name (ie DumboBot,LabBot, etc) and use getattr() to call the class in the
        # imported module
        obj = None
        try:
            obj = getattr(module, self.robot_class_name)()
        except Exception as e:
            err_msg = f"Unable to get attributes on {module}.{self.robot_class_name} please ensure that your ~/robotConfig exists, and contents are correct. robot name={self.raw_name} Exception: {e}"
            log.error(err_msg)
            raise Exception(err_msg)

        # return getattr(module, self.robot_class_name)()
        return obj


def get_bot(name: str) -> ConfigBasedCommandRobot:
    """getBot will get the configs and create the bot for the supplied name.
    If no name is supplied, then it will query the filesystem for ~/robotConfig and
    read the robot name from the file (yaml extension not necessary) Bot is assumed at the end of the name.
    Eventually this method should be able to query the RoboRio itself for the robot name but this still still unimplmented."

        Returns:
            ConfigBasedCommandRobot: Instance of a name of the robot like green, dumbo, teapot, lab, etc
    """
    ic(name)
    return BotFactory(name).get_robot()



import rev

def configureSparkMaxCanRates(motor : rev.CANSparkMax, faultRateMs = 50, motorTelmRateMs = 50, motorPosRateMs = 50, analogRateMs=1833, altEncoderRateMs = 1050, dutyCycleEncRateMs = 2150, dutyCycleEncVelRateMs = 3150):
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus0, faultRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus1, motorTelmRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus2, motorPosRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus3, analogRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus4, altEncoderRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus5, dutyCycleEncRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus6, dutyCycleEncVelRateMs)
    motor.setPeriodicFramePeriod(rev.CANSparkLowLevel.PeriodicFrame.kStatus7, 500) #Unknown frame type? default 250ms prob not important?
    



"""
Contains helpers to create various sensor types
"""

from wpilib import DigitalInput as di
import logging
import navx
import wpilib
import traceback

log = logging.getLogger("SensorFactory")

def gyroFactory(descp):
    """
    Creates gyros from a gyro descp
    """
    try:
        if "navx" == descp["type"]:
            """
            Supports spi and i2c with default values. More can be added as needed.
            """
            method = descp["method"]
            if method == "spi":
                return navx.AHRS.create_spi()
            if method == "i2c":
                return navx.AHRS.create_i2c()
            #invalid method, thorw a fit
            raise ValueError(f"{method} method is invalid")

    except Exception as e:
        logging.error("Failed to create gyro for %s. Error %s",descp, e)
    return None

def breaksensorFactory(descp):
    """
    Creates break sensors from a break sensor descp
    """
    try:
        if "RIODigitalIn" in descp["type"]:
            return di(descp["channel"])

    except Exception as e:
        logging.error("Failed to create IR Break sensor for %s. Error %s", descp, e)
    return None


def dutyCycleEncoderFactory(config: dict) -> wpilib.DutyCycleEncoder:
    """
        config format:
        type: "sensor.wpilib.DutyCycleEncoder"
        channel: digital IO channel (i.e. 0). required
        offset: sensor offset in units (i.e. 180.0 for 180 degrees). default 0.0
        unitsPerRotation: units per rotation (i.e 360.0 for degrees). default 1.0
        minDutyCycle: see setDutyCycleRange. default 0
        maxDutyCycle: see setDutyCycleRange. default 1.0
    """
    channel = config["channel"] #raise value error if not present
    offset = config.get("offset", 0.0)
    upr = config.get("unitsPerRotation", 1.0)
    minDuty = config.get("minDutyCycle", 0.0)
    maxDuty = config.get("maxDutyCYcle", 1.0)
    log.info(f"Creating DutyCycleEncoder")
    log.info(f"Channel {channel}, UnitsPerRotation {upr}, offset {offset}, minDutyCycle {minDuty} maxDutyCycle {maxDuty}")
    encoder = wpilib.DutyCycleEncoder(config["channel"])
    encoder.setDistancePerRotation(upr)
    encoder.setPositionOffset(offset)
    encoder.setDutyCycleRange(minDuty, maxDuty)
    return encoder

def create(compType: str, config: dict):
    try:
        match compType:
            case "wpilib.DutyCycleEncoder":
                log.info("Creating DutyCycleEndoer")
                return dutyCycleEncoderFactory(config)
            case _:
                log.error(f"Unsupported Sensor {compType} for {config}")
    except Exception as e:
        log.error(f"Failed to create sensor {compType} for {config}")
        log.error(f"Error caught: {e}")
        traceback.print_exception(e)

        if config.get("required", False):
            raise RuntimeError(f"Failed to create required component {compType} for {config}")
    
    return None



# -*- coding: utf-8 -*-

import rev
import ctre
import logging as log
from raptacon3200.utils.unit_enums import positionUnits, velocityUnits


def createMotor(motorDescp, motors={}):
    """This is where all motors are set up.
    Motors include CAN Talons, CAN Talon Followers, CAN Talon FX, CAN Talon FX Followers, and SparkMax and its follower.
    Not all are functional, it's up to you to find out. Good luck!"""
    if motorDescp["type"] == "CANTalonSRX":
        # if we want to use the built in encoder set it here
        if ("pid" in motorDescp) and motorDescp["pid"] != None:
            motor = WPI_TalonSRXFeedback(motorDescp)
            motor.setupPid()
        else:
            motor = ctre.WPI_TalonSRX(motorDescp["channel"])
        setTalonSRXCurrentLimits(motor, motorDescp)
        motors[str(motorDescp["channel"])] = motor

    elif motorDescp["type"] == "CANTalonSRXFollower":
        """This is where we set up Talon SRXs over CAN"""
        motor = ctre.WPI_TalonSRX(motorDescp["channel"])
        motor.set(mode=ctre.ControlMode.Follower, value=motorDescp["masterChannel"])
        setTalonSRXCurrentLimits(motor, motorDescp)
        motors[str(motorDescp["channel"])] = motor

    elif motorDescp["type"] == "CANTalonFX":
        """This is where CANTalon FXs are set up"""
        if ("pid" in motorDescp) and motorDescp["pid"] != None:
            motor = WPI_TalonFXFeedback(motorDescp)
            motor.setupPid()
        else:
            motor = ctre.WPI_TalonFX(motorDescp["channel"])
        setTalonFXCurrentLimits(motor, motorDescp)

    elif motorDescp["type"] == "CANTalonFXFollower":
        """This is where CANTalon FX Followers are set up"""
        motor = ctre.WPI_TalonFX(motorDescp["channel"])
        motor.set(
            mode=ctre.TalonFXControlMode.Follower, value=motorDescp["masterChannel"]
        )
        motors[str(motorDescp["channel"])] = motor
        setTalonFXCurrentLimits(motor, motorDescp)

    elif motorDescp["type"] == "SparkMax":
        """This is where SparkMax motor controllers are set up"""
        motorDescp["motorType"] = getattr(
            rev.CANSparkMax.MotorType, motorDescp["motorType"]
        )

        if "pid" in motorDescp and motorDescp["pid"] != None:
            motor = SparkMaxFeedback(motorDescp, motors)
            motor.setupPid()
        else:
            motor = rev.CANSparkMax(motorDescp["channel"], motorDescp["motorType"])

        motors[str(motorDescp["channel"])] = motor
        setREVCurrentLimits(motor, motorDescp)

    elif motorDescp["type"] == "SparkMaxFollower":
        """This is where SparkMax followers are set up
        For masterChannel, use a motor object. MASTER MUST BE A "CANSparkMax" because blame rev
        """
        motorDescp["motorType"] = getattr(rev.MotorType, motorDescp["motorType"])
        motor = SparkMaxFeedback(motorDescp, motors)
        motor.follow(
            motors.get(str(motorDescp["masterChannel"])), motorDescp["inverted"]
        )
        setREVCurrentLimits(motor, motorDescp)

    else:
        print("Unknown Motor")
        return None

    if "inverted" in motorDescp:
        motor.setInverted(motorDescp["inverted"])

    return motor


def setTalonFXCurrentLimits(motor, motorDescp):
    """
    Sets current limits based off of "currentLimits"
    in your motor and config of choice. Must be a Talon FX motor controller
    In currentLimits, you need currentLimit, triggerThresholdCurrent, and triggerThresholdTime.
    """
    if "currentLimits" in motorDescp:
        currentLimits = motorDescp["currentLimits"]
        currentLimit = currentLimits["currentLimit"]
        triggerThresholdCurrent = currentLimits["triggerThresholdCurrent"]
        triggerThresholdTime = currentLimits["triggerThresholdTime"]
        statorCurrentConfig = ctre.StatorCurrentLimitConfiguration(
            True, currentLimit, triggerThresholdCurrent, triggerThresholdTime
        )
        supplyCurrentConfig = ctre.SupplyCurrentLimitConfiguration(
            True, currentLimit, triggerThresholdCurrent, triggerThresholdTime
        )
        motor.configStatorCurrentLimit(statorCurrentConfig)
        motor.configSupplyCurrentLimit(supplyCurrentConfig)


def setTalonSRXCurrentLimits(motor, motorDescp):
    """
    Sets current limits based off of "currentLimits"
    in your motor and config of choice. Must be a Talon SRX motor controller
    In currentLimits, you need absMax, absMaxTimeMs, maxNominal.
    """
    if "currentLimits" in motorDescp:
        currentLimits = motorDescp["currentLimits"]
        absMax = currentLimits["absMax"]
        absMaxTimeMs = currentLimits["absMaxTimeMs"]
        nominalMaxCurrent = currentLimits["maxNominal"]
        motor.configPeakCurrentLimit(absMax, 10)
        motor.configPeakCurrentDuration(absMaxTimeMs, 10)
        motor.configContinuousCurrentLimit(nominalMaxCurrent, 10)
        motor.enableCurrentLimit(True)


def setREVCurrentLimits(motor, motorDescp):
    """
    Sets current limits based off of "currentLimits"
    in your motor and config of choice. Must be a REV motor controller
    In currentLimits, you need freeLimit, stallLimit, stallLimitRPM and secondaryLimit
    """
    if "currentLimits" in motorDescp:
        currentLimits = motorDescp["currentLimits"]
        freeLimit = currentLimits["freeLimit"]
        stallLimit = currentLimits["stallLimit"]
        limitRPM = currentLimits["stallLimitRPM"]
        secondaryLimit = currentLimits["secondaryLimit"]
        motor.setSecondaryCurrentLimit(secondaryLimit)
        motor.setSmartCurrentLimit(stallLimit, freeLimit, limitRPM)


class WPI_TalonSRXFeedback(ctre.WPI_TalonSRX):  # ctre.wpi_talonsrx.WPI_TalonSRX
    """
    Class used to setup TalonSRX motors if there are PID setting for it
    """

    def __init__(self, motorDescription):
        ctre.WPI_TalonSRX.__init__(self, motorDescription["channel"])
        self.motorDescription = motorDescription
        self.pid = None

    def setupPid(self, motorDescription=None):
        """Sets up PID based on dictionary motorDescription['pid'].
        This dictionary must contain controlType, feedbackDevice, sensorPhase, kPreScale, and P, I, D and F.
        """
        if not motorDescription:
            motorDescription = self.motorDescription
        if not "pid" in self.motorDescription:
            print("Motor channel %d has no PID" % (self.motorDescription["channel"]))
            return
        self.pid = self.motorDescription["pid"]

        # Takes a str and converts it to a ctre enum
        self.controlType = self.pid["controlType"]
        if self.controlType == "Position":
            self.controlType = ctre.ControlMode.Position
        elif self.controlType == "Velocity":
            self.controlType = ctre.ControlMode.Velocity

        self.configSelectedFeedbackSensor(
            ctre.FeedbackDevice(self.pid["feedbackDevice"]), 0, 10
        )
        self.setSensorPhase(self.pid["sensorPhase"])
        self.ControlType = self.pid["controlType"]
        self.kPreScale = self.pid["kPreScale"]

        # /* set the peak, nominal outputs, and deadband */
        self.configNominalOutputForward(0, 10)
        self.configNominalOutputReverse(0, 10)
        self.configPeakOutputForward(1, 10)
        self.configPeakOutputReverse(-1, 10)

        self.configVelocityMeasurementPeriod(ctre.VelocityMeasPeriod(1), 10)
        # /* set closed loop gains in slot0 */
        self.config_kF(0, self.pid["kF"], 10)
        self.config_kP(0, self.pid["kP"], 10)
        self.config_kI(0, self.pid["kI"], 10)
        self.config_kD(0, self.pid["kD"], 10)

    def set(self, speed):
        if self.pid != None:
            return ctre.WPI_TalonSRX.set(self, self.controlType, speed * self.kPreScale)
        else:
            return self.set(speed)


class WPI_TalonFXFeedback(ctre.WPI_TalonFX):
    def __init__(self, motorDescription):
        """Sets up the basic Talon FX with channel of motorDescription['channel']. Doesn't set up pid."""
        ctre.WPI_TalonFX.__init__(self, motorDescription["channel"])
        self.motorDescription = motorDescription
        self.pid = None
        if self.motorDescription["type"] == "CANTalonFXFollower":
            self.controlType = ctre.TalonFXControlMode.Follower
        else:
            self.controlType = ctre.TalonFXControlMode.PercentOutput

    def setupPid(self, motorDescription=None):
        """Sets up pid based on the dictionary motorDescription['pid']
        (Must contain kP, kI, kD, kF, controlType, sensorPhase (boolean), kPreScale, feedbackDevice)
        """
        if not motorDescription:
            motorDescription = self.motorDescription
        if not "pid" in self.motorDescription:
            log.error(
                "Motor channel " + str(self.motorDescription["channel"]) + " has no PID"
            )
            return
        self.pid = self.motorDescription["pid"]

        # Takes a str and converts it to a ctre enum for controltype.
        self.controlType = self.pid["controlType"]
        if self.controlType == "Position":
            self.controlType = ctre.TalonFXControlMode.Position
        elif self.controlType == "Velocity":
            self.controlType = ctre.TalonFXControlMode.Velocity
        elif self.controlType == "PercentOutput":
            # This is so that we can initialize a motor as percentoutput but also use an encoder
            self.controlType = ctre.ControlMode.PercentOutput
        else:
            log.error("Unrecognized control type: " + str(self.ControlType))

        if self.pid["feedbackDevice"] == "IntegratedSensor":
            # This is the feedbackDevice for TalonFXs for the integrated sensor
            feedbackDevice = ctre.FeedbackDevice.IntegratedSensor
        else:
            log.error("Unrecognized feedbackDevice " + str(self.pid["feedbackDevice"]))
            return

        self.configSelectedFeedbackSensor(feedbackDevice, 0, 10)
        self.setSensorPhase(self.pid["sensorPhase"])
        self.kPreScale = self.pid["kPreScale"]

        # /* set the peak, nominal outputs, and deadband */
        self.configNominalOutputForward(0, 10)
        self.configNominalOutputReverse(0, 10)
        self.configPeakOutputForward(1, 10)
        self.configPeakOutputReverse(-1, 10)
        self.configVelocityMeasurementPeriod(ctre.SensorVelocityMeasPeriod(1), 10)
        # /* set closed loop gains in slot 0 */
        self.config_kF(0, self.pid["kF"], 10)
        self.config_kP(0, self.pid["kP"], 10)
        self.config_kI(0, self.pid["kI"], 10)
        self.config_kD(0, self.pid["kD"], 10)

        self.sensorCollection = self.getSensorCollection()

    def setBraking(self, braking: bool):
        if braking:
            self.setNeutralMode(ctre.NeutralMode.Brake)
        else:
            self.setNeutralMode(ctre.NeutralMode.Coast)

    def resetPosition(self):
        self.sensorCollection.setIntegratedSensorPosition(0)

    def getPosition(self, pidId, units: positionUnits):
        """
        pidId: The ID of the pid config
        (0 for primary, 1 for auxilliary)

        Returns the integrated sensor's current position in
        encoder ticks (2048 per 1 rotation)
        """
        if units == positionUnits.kEncoderTicks:
            self.position = self.sensorCollection.getIntegratedSensorPosition()
        elif units == positionUnits.kRotations:
            self.position = self.sensorCollection.getIntegratedSensorPosition() / 2048
        else:
            log.error("Unrecognized units: " + str(units))
            return "Unrecognized unit"

        return self.position

    def getVelocity(self, pidId, units: velocityUnits):
        """
        pidId: The ID of the pid config
        (0 for primary, 1 for auxilliary)
        units:

        Returns the integrated sensor's current velocity in
        encoder ticks / 100 ms (2048 encoder ticks per 1 rotation)
        """
        if units == velocityUnits.kEncoderTicksPer100:
            self.velocity = self.sensorCollection.getIntegratedSensorVelocity()
        elif units == velocityUnits.kRPS:
            self.velocity = (
                10 * self.sensorCollection.getIntegratedSensorVelocity()
            ) / 2048
        elif units == velocityUnits.kRPM:
            self.velocity = (
                self.sensorCollection.getIntegratedSensorVelocity() / 2048
            ) / 600
        else:
            log.error("Unrecognized units: " + str(units))
            return "Unrecognized unit"

        return self.velocity

    def set(self, speed):
        """
        Overrides the default set() to allow for control using the pid loop
        """
        if self.pid:
            return ctre.WPI_TalonFX.set(self, self.controlType, speed * self.kPreScale)
        else:
            return ctre.WPI_TalonFX.set(self, speed)


class SparkMaxFeedback(rev.CANSparkMax):
    """
    Class used to setup SparkMax motor if there are PID settings for it - MUST CALL setupPID
    if you don't want it to crash. Great design decision on my part.
    """

    def __init__(self, motorDescription, motors):
        self.motorDescription = motorDescription
        self.motorType = self.motorDescription["motorType"]

        rev.CANSparkMax.__init__(self, self.motorDescription["channel"], self.motorType)
        self.setInverted(self.motorDescription["inverted"])
        self.motors = motors
        self.coasting = False
        # Generally just a way to overwrite previous settings on any motor controller - We don't brake often.
        if (
            "IdleBrake" in self.motorDescription.keys()
            and self.motorDescription["IdleBrake"] == True
        ):
            self.setIdleMode(rev.CANSparkMax.IdleMode.kBrake)
        else:
            self.setIdleMode(rev.CANSparkMax.IdleMode.kCoast)

    def setupPid(self):
        """Sets up the PIDF values and a pidcontroller to use to control the motor using pid."""
        if not "pid" in self.motorDescription:
            print("Motor channel %f has no PID", (self.motorDescription["channel"]))
            return
        self.pid = self.motorDescription["pid"]
        pid = self.pid
        self.ControlType = pid["controlType"]

        # Turns strings from pid dictionary in config into enums from rev library for control type
        if self.ControlType == "Position":
            self.ControlType = rev.CANSparkMaxLowLevel.ControlType.kPosition
        elif self.ControlType == "Velocity":
            self.ControlType = rev.CANSparkMaxLowLevel.ControlType.kVelocity
        else:
            print("Unrecognized control type: ", self.ControlType)

        # If coastOnZero is true, when set() is called with a speed of 0, we will use DutyCycle
        # And let the motor spin down by itself. (demonstrated in coast and stopcoast methods and within set())
        if "coastOnZero" in self.pid and self.pid["coastOnZero"] == True:
            self.coastOnZero = True
        else:
            self.coastOnZero = False

        self.prevControlType = self.ControlType

        self.encoder = self.getEncoder()
        self.kPreScale = pid[
            "kPreScale"
        ]  # Multiplier for the speed - lets you stay withing -1 to 1 for input but different outputs to pidController
        self.PIDController = self.getPIDController()  # creates pid controller

        # Sets PID(F) values
        self.PIDController.setP(
            pid["kP"], pid["feedbackDevice"]
        )  # pid['feedbackDevice'] is a slot for PID(F) configs. They range from 0-3.
        self.PIDController.setI(pid["kI"], pid["feedbackDevice"])
        self.PIDController.setD(pid["kD"], pid["feedbackDevice"])
        self.PIDController.setFF(pid["kF"], pid["feedbackDevice"])

        # Configures output range - that's what Spark Maxes accept
        self.PIDController.setOutputRange(-1, 1, pid["feedbackDevice"])
        self.PIDController.setReference(0, self.ControlType, pid["feedbackDevice"])

    def setControlType(self, type: str):
        """
        Takes str type as argument, currently accepts Position, Velocity and Duty Cycle.
        More can be added as necessary, following previous syntax in this method.
        """
        if type == "Position":
            self.ControlType = rev.CANSparkMax.ControlType.kPosition
        elif type == "Velocity":
            self.ControlType = rev.CANSparkMax.ControlType.kVelocity
        elif type == "Duty Cycle":
            self.ControlType = rev.CANSparkMax.ControlType.kDutyCycle
        else:
            print("Unrecognized control type: ", self.ControlType)

    def coast(self):
        """
        Stores the current control type, moves to Duty Cycle, sets to 0.
        """
        if self.coasting:
            return
        self.coasting = True
        self.prevControlType = self.ControlType
        self.setControlType("Duty Cycle")
        self.PIDController.setReference(0, self.ControlType, self.pid["feedbackDevice"])

    def stopCoast(self):
        """
        Restores previous control type. Whatever it was.
        """
        if self.coasting:
            self.ControlType = self.prevControlType
        self.coasting = False

    def set(self, speed):
        """
        Overrides the default set() to allow for control using the pid loop
        """
        if self.coastOnZero and speed == 0:
            self.coast()
        else:
            self.stopCoast()
        return self.PIDController.setReference(
            speed * self.pid["kPreScale"], self.ControlType, self.pid["feedbackDevice"]
        )



import importlib
import inspect
from os import path


def botFinder() -> str:
    """Looks in call stack for first robot.py and returns the fully qualified path to that file

    Returns:
        str: Fully qualified path to robot.py from call stack (minus robot.py)
    """
    stack = inspect.stack()
    for f_idx, f_info in enumerate(stack):
        # print(f"robot_info stack {f_idx} {f_info}")
        frame, filename, line_number, function_name, lines, index = stack[f_idx]
        # print(
        #     f"looking for {f_idx} {f_info} {frame}, {filename}, {line_number}, {function_name}, {lines}, {index}"
        # )
        if "robot.py" in filename:
            print(f"FOUND!!! {filename}")
            return path.dirname(filename)
    return None



import logging
from raptacon3200.utils import motorHelper
from raptacon3200.bot_factory import sensorFactory

# TODO Delet me
logging.basicConfig(level=logging.DEBUG)

log = logging.root.getChild("HardwareFactory")
# log.setLevel(logging.DEBUG)


class HardwareFactory(object):
    """
    Hardware factory for instantiating hardware components in a config
    """

    def __new__(cls):
        if not hasattr(cls, "instance"):
            log.debug("New instance created")
            cls.instance = super(HardwareFactory, cls).__new__(cls)
        else:
            log.debug("HardwareFactory alread created returning singleton")
        return cls.instance

    def __init__(self):
        log.debug("New HardwareFactory")
        self.config = {}
        self.components = {}

    def addConfig(self, subsystem: str, config: dict):
        """
        Adds a new config. Replaces or updates exsisting components.
        """
        log.debug(f"Adding config {subsystem}.{config}")

        if subsystem == None:
            log.debug("Adding to root")
            subsystem = "root"

        # TODO add logic to check if a HW instance was alread created. If so need to log a warning. Crashing for now
        if subsystem in self.config:
            assert False, "Subsystem updatinng and saftey not done"
        self.config[subsystem] = config

        for name, value in self.config[subsystem].items():
            # create the hardware components at config time to catch errors early
            if isinstance(value, dict) and "type" in value:
                log.debug(f"Creating {subsystem}.{name} type {value['type']}")
                self.createHardwareComponent(subsystem, name, value)
            else:
                log.info(f"{subsystem}.{name} is not hardware")

    def getConfig(self, subsystem: str, name: str):
        """
        returns the config for a given subsystem / name.
        If subsystem = None, searchs for the first name that matches.
        Returns None on failure to find config
        """
        if not subsystem:
            for subsystem in self.config:
                if name in self.config[subsystem]:
                    return self.config[subsystem]

        elif subsystem in self.config:
            if name in self.config[subsystem]:
                return self.config[subsystem][name]
        log.info(f"getConfig({subsystem}, {name}) failed to find config")
        return None

    def getHardwareComponent(self, subsystem: str, name: str):
        """
        Returns an instantiated component. Return None if does not exsist
        """
        if subsystem == None:
            subsystem = "any"
        log.debug(f"Geting component {subsystem}.{name}")

        component = None
        for ss in self.components:
            if ss == subsystem or subsystem == "any":
                if name in self.components[ss]:
                    component = self.components[ss][name]
                    log.debug(f"Got {component} from {ss}.{name}")
                    return component

        log.debug(f"{subsystem}.{name} not created. Creating")
        if self.createHardwareComponent(subsystem, name):
            log.error("Not finished")
            return self.components[name]
        else:
            log.error(f"Failed to find component {name}")
            return None

    def createHardwareComponent(self, subsystem: str, name: str, config: dict = None):
        """
        Creates a component from a dictory.
        """

        if not config:
            config = self.getConfig(subsystem, name)

        config = config.copy()
        try:
            subtype = config["type"].split(".", 1)[0]
            compType = config["type"].split(".", 1)[1]
            log.info(f"Creating component {subsystem}.{name} as {subtype}.{compType}")
        except Exception as e:
            log.error(
                f"**** {config['type']} is not in the format subtype.type. i.e. motor.SparkMax) for {subsystem} {name}"
            )
            log.debug(e)
            if "required" in config and config["required"]:
                raise RuntimeError(
                    f"Failed to create required hardware {config['type']} in {subsystem} {name}"
                )

            return None

        # clean up components
        if not subsystem in self.components:
            self.components[subsystem] = {}
        if not name in self.components[subsystem]:
            self.components[subsystem][name] = {}

        # use short name without subtype
        config["type"] = compType
        component = None
        match subtype:
            case "motor":
                component = motorHelper.createMotor(config)
            case "sensor":
                component = sensorFactory.create(compType, config)
            case _:
                log.error(
                    f"{subtype} is not a valid subtype. Support for matching any factory not done"
                )
        if component:
            # add names for later use
            self.components[subsystem][name] = component
        elif "required" in config and config["required"]:
            raise RuntimeError(
                f"Failed to create {subsystem}.{name} {subtype}.{compType} and is required"
            )

        return component


def getHardwareFactory() -> HardwareFactory:
    return HardwareFactory()



from raptacon3200 import yaml  # So why not pyyaml?
import logging
from pprint import pprint
import os
import sys
from pathlib import Path
from raptacon3200.bot_factory import hardwareFactory
from raptacon3200.utils import botFinder as utils
import importlib

# Easy Line Debugging
# https://github.com/gruns/icecream
from icecream import ic

log = logging.getLogger("configMapper")


class ConfigMapper(object):
    def __init__(self, filename, configDir):
        """
        Initlizes the config off a tree of yamls.
        "/" holds global config
        "<subsystem>" holds configs for subsystems
        configDir points to folder with configs. Future work make it take a list and search.
        """
        self.configDir = configDir
        initialData = self.__loadFile(filename)
        log.debug("Intial data %s", initialData)
        self.subsystems = self.__convertToSubsystems(initialData, "/")
        self.hwFact = hardwareFactory.getHardwareFactory()

        for key, subsystem in self.subsystems.items():
            self.hwFact.addConfig(key, subsystem)

    def getSubsystem(self, subsystemName):
        """
        returns the complete config for specified subsystem or none if not
        found
        """
        # gives the values.

        if subsystemName not in self.subsystems:
            raise RuntimeError(f"Failed to find {subsystemName} in config")

        subsystem = self.subsystems[subsystemName]

        if "subsystem" not in subsystem:
            raise RuntimeError(
                f"Failed to find `subsystem` in {subsystemName}. Found {subsystem}"
            )

        # load and create subsystem
        subClass = importClassFromModule(subsystem["subsystem"])

        if subClass == None:
            if subsystem["required"]:
                raise RuntimeError(
                    f"Failed to create {subsystemName} from {subsystem['subsystem']} and is required"
                )
            else:
                log.warn(
                    f"Failed to create {subsystemName} from {subsystem['subsystem']} and is not required"
                )
                return None
        subsystem["configMapper"] = self

        return subClass(**subsystem)

    def getSubsystems(self):
        subsystems = list(self.subsystems.keys())
        return subsystems

    def getGroupDict(self, subsystem, groupName, name=None):
        """
        returns a dictonary with data from a subsystem matching the groups and
        if given then name.
        i.e.
        if you have a lifter with
        lifter.sensors.groups=sensor
        lifter.intakeMotors.groups=motors
        lifter.beltMotors.groups=motors
        calling getTypeDict("lifter", "motors")
        returns all motors in intakeMotors and beltMotors

        calling getTypeDict("lifter", "motors", "beltMotors")
        returns all motors in beltMotors

        calling getTypeDict("lifter", "sensor")
        returns all sensors
        """
        data = self.getSubsystem(subsystem)
        data = self.__getGroups(data, groupName, name)
        if "groups" in data:
            data.pop("groups")

        return data

    def getTypesDict(self, subsystem, typeNames, name=None):
        """
        returns a dictonary with data from a subsystem matching the type(s) and
        Once a type is found in an entry it is not searched any deeper
        """
        if not isinstance(typeNames, list):
            typeNames = [typeNames]
        data = self.getSubsystem(subsystem)
        data = self.__getTypes(data, typeNames, name)
        return data

    def __getGroups(self, data, groupName, name):
        """
        internal call, recusivley search the data for entries with
        groups in "types" and if a name is given only types inside key with name
        """

        retVal = {}
        for key in data:
            if isinstance(data[key], dict):
                recusiveDict = self.__getGroups(data[key], groupName, name)
                retVal.update(recusiveDict)

            if isinstance(data[key], dict) and "groups" in data[key]:
                if name and not key == name:
                    continue

                if groupName in data[key]["groups"]:
                    retVal.update(data[key])
        return retVal

    def __getTypes(self, data, typeNames, name):
        """
        internal call, recusivley search the data for entries with
        typeName in ["types"] and if a name is given only types
        inside key with name
        """
        retVal = {}
        for key in data:
            if isinstance(data[key], dict):
                recusiveDict = self.__getTypes(data[key], typeNames, name)
                retVal.update(recusiveDict)

            if isinstance(data[key], dict) and "type" in data[key]:
                if name and not key == name:
                    continue

                if data[key]["type"] in typeNames:
                    retVal[key] = data[key]
        return retVal

    def __loadFile(self, filename):
        """
        Loads a yaml or yml file and returns the contents as dictionary
        """
        with open(self.configDir + os.path.sep + filename) as file:
            values = yaml.load(file, yaml.FullLoader)
            return values

    def __convertToSubsystems(self, inputData, defSubsystem):
        """
        Takes a dictionary and searchs for subsystem types to create leafs of a new tree.
        Loads files as "file" is encountered
        """
        print(inputData)

        if "subsystems" not in inputData:
            log.error("No Subsystems included in config")
        subsystems = inputData["subsystems"]
        processedData = {}

        for name, subsystem in subsystems.items():
            processedData[name] = {}
            # add required data defaults
            processedData[name]["required"] = True

            for key, value in subsystem.items():
                # if file, load file and walk
                if key == "file":
                    file = "subsystems" + os.sep + value
                    # TODO special handling for other types. Yaml only supproted type
                    if "type" in subsystem and not subsystem["type"] == "yaml":
                        log.error("Unknown file type fileType. Trying Yaml")
                    log.info("Loading %s into entry %s", file, key)
                    try:
                        data = self.__loadFile(file)
                    except FileNotFoundError as e:
                        log.error(
                            f"Failed to find file {file} for subsystem {subsystem}"
                        )
                        data = {}
                        data["error"] = e
                    # Flatten the root node of newly loaded yaml file.
                    processedData[name].update(data)

                # if subsystem, walk subsystem
                # TODO add nested subsystems
                """
                if "subsystem" in subsystem[key] and isinstance(subsystem[key], dict):
                    log.info("Walking subsystem")
                    # make a new subsystem
                    print(subsystem[key])
                    print(subsystem[key]["subsystem"])
                    processedData[subsystem[key]["subsystem"]] = self.__convertToSubsystems(inputData[key], inputData[key]["subsystem"])
                """

                # skip special meanings
                if key == "type" and "file" in subsystem:
                    continue

                if key == "file":
                    continue

                # copy field over if no special processing
                processedData[name][key] = subsystem[key]

        return processedData


def importClassFromModule(name: str, base: str = "subsystems"):
    """
    Imports a class from a given module

    Args:
        name (str): Name of class in module.path.Class format
        base (str, optional): base module to load from. Defaults to "subsystems".

    Returns:
        _type_: Class or None
    """
    name = base + "." + name
    moduleName, className = name.rsplit(".", 1)

    log.info(f"Loading {name} as module {moduleName} and class {className}")

    module = importlib.import_module(moduleName)
    try:
        return getattr(module, className)
    except:
        log.warning(f"Failed to find {className} in {module}")
        return None


def findConfig(defaultConfig="greenbot.yml", configPath=None) -> tuple[dict, str]:
    """
    Will determine the correct yml file for the robot.
    Please run 'echo (robotCfg.yml) > robotConfig' on the robot.
    This will tell the robot to use robotCfg file remove the () and use file name file.
    Files should be configs dir
    """
    if not configPath:
        ic("In not having configPath")
        configPath = utils.botFinder() + os.path.sep + "configs" + os.path.sep
        # configPath = (
        #     os.path.dirname(__file__)
        #     + os.path.sep
        #     + ".."
        #     + os.path.sep
        #     + "configs"
        #     + os.path.sep
        # )
    home = str(Path.home()) + os.path.sep
    ic(home)
    robotConfigFile = home + "robotConfig"
    ic(robotConfigFile)

    if not os.path.isfile(robotConfigFile):
        log.error("Could not find %s. Using default", robotConfigFile)
        robotConfigFile = configPath + "default"
    try:
        file = open(robotConfigFile)
        configFileName = file.readline().strip()
        file.close()
        configFile = configPath + configFileName
        if os.path.isfile(configFile):
            log.info("Using %s config file", configFile)
            return configFileName, configPath
        log.error("No config? Can't find %s", configFile)
        log.error("Using default %s", defaultConfig)
    except Exception as e:
        log.error("Could not find %s", robotConfigFile)
        log.error(e)
        log.error("Please run `echo <robotcfg.yml> > ~/robotConfig` on the robot")
        log.error("Using default %s", defaultConfig)

    ic(defaultConfig)
    ic(configPath)
    return defaultConfig, configPath


if __name__ == "__main__":

    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

    mapper = ConfigMapper("greenBot.yml", "configs")
    print("Subsystem driveTrain:", mapper.getSubsystem("drivetrain"))

    print("driveTrain Motors")
    pprint(mapper.getGroupDict("driveTrain", "motors"))

    print("Shooter motors:")
    pprint(mapper.getGroupDict("shooter", "motors", "loaderMotors"))

    print("All motors:")
    mapper.getGroupDict("/", "motors")
    # print()
    pprint(mapper.getGroupDict("/", "motors"))

    print("CANTalonFXFollower motors:")
    data = mapper.getTypesDict("/", "CANTalonFXFollower")
    # print()
    pprint(data)

    compatTest = ["Dog", "all", "doof", "minibot", "DOOF"]
    for item in compatTest:
        compat = mapper.checkCompatibilty(item)
        print(f"{item} is {compat}")

    print("Subsystems: ", mapper.getSubsystems())
